# __version__.py
# autogenerated by poetry-hooks 0.1.0

__version__ = "0.1.0a0"
__title__ = "caldera"
__authors__ = ["Justin Vrana <justin.vrana@gmail.com>"]
__repository__ = ""
__homepage__ = "http://www.github.com/jvrana/caldera"
__description__ = ""
__maintainers__ = ""
__readme__ = ""
__license__ = ""
