""""""
from ._version import __authors__  # noqa: F401
from ._version import __homepage__  # noqa: F401
from ._version import __repository__  # noqa: F401
from ._version import __title__  # noqa: F401
from ._version import __version__  # noqa: F401
from caldera._setup import setup as _setup

_setup()
